from ..client.game_client import GameClient
import asyncio
from loguru import logger
from typing import Any



class Map(GameClient):
    
    
    async def get_map_chunk(self, kingdom, x, y):
        
        try:
            
            await self.send_json_message("gaa",{
                "KID": kingdom,
                "AX1": x,
                "AY1": y,
                "AX2": x + 12,
                "AY2": y + 12
            })
            
            
            
        except Exception as e:
            logger.error(e)
            return False
    
    
    async def get_closest_npc(
        self,
        kingdom: int,
        npc_type: int,
        min_level: int = 1,
        max_level: int = -1,
        owner_id: int = -1,
        sync: bool = True
        ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "fnm",
                {
                    
                    "T": npc_type,
                    "KID": kingdom,
                    "LMIN": min_level,
                    "LMAX": max_level,
                    "NID": owner_id
                }
            )
            if sync:
                response = await self.wait_for_response("fnm")
                return response
            return True
            
        except Exception as e:
            logger.error(e)
            return False
        
        

    
    def _spiral_coords(self, max_radius: int, cx: int, cy: int):
        step = 13
        seen = set()
        for radius in range(step, max_radius + 1, step):

            for x in range(cx - radius, cx + radius + 1, step):
                pt = (x, cy - radius)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
           
            for y in range(cy - radius, cy + radius + 1, step):
                pt = (cx + radius, y)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
            
            for x in range(cx + radius, cx - radius - 1, -step):
                pt = (x, cy + radius)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
           
            for y in range(cy + radius, cy - radius - 1, -step):
                pt = (cx - radius, y)
                if pt not in seen:
                    seen.add(pt)
                    yield pt
     
    
    
    
    async def map_scanner(
        self,
        kingdom: int,
        max_radius: int,
        castle_x: int,
        castle_y: int,
        handler: Any
        ):

        
        processing_task = asyncio.create_task(
            self.process_data("gaa", handler, interval=0.5)
        )
        try:
            logger.info("Scanning...")
            
            for x, y in self._spiral_coords(max_radius=max_radius, cx=castle_x, cy=castle_y):
                await self.get_map_chunk(kingdom, x, y)
                await asyncio.sleep(0.01)
            logger.info("Complete")
        finally:
            
            processing_task.cancel()
            
            
     
     
            
    async def find_by_name(
        self,
        user_name: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("wsp", {"PN": user_name})
            if sync:
                response = await self.wait_for_response("wsp")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    async def get_npc_target_infos(
        self,
        tx: int,
        ty: int,
        sx: int,
        sy: int,
        kid: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "adi",
                {
                    "TX": tx,
                    "TY": ty,
                    "SX": sx,
                    "SY": sy,
                    "KID": kid
                }
            )
            if sync:
                response = await self.wait_for_response("adi")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        




    async def get_pvp_target_infos(
        self,
        tx: int,
        ty: int,
        sx: int,
        sy: int,
        kid: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "aci",
                {
                    "TX": tx,
                    "TY": ty,
                    "SX": sx,
                    "SY": sy,
                    "KID": kid
                }
            )
            if sync:
                response = await self.wait_for_response("aci")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
        
    def periodic_map_scanner(
        self,
        kingdom: int,
        max_radius: int,
        castle_x: int,
        castle_y: int,
        handler: Any,
        interval: float,
        run_immediately: bool = True
    ):
        loop = asyncio.get_running_loop()
        
        async def _scan_and_reschedule():
            
            try:
                await self.map_scanner(kingdom, max_radius, castle_x, castle_y, handler)
            
            except Exception as e:
                logger.error(e)
            
            finally:
                loop.call_later(interval, lambda: loop.create_task(_scan_and_reschedule()))
                
            if run_immediately:
                loop.create_task(_scan_and_reschedule())
            
            else:
                loop.call_later(interval, lambda: loop.create_task(_scan_and_reschedule()))


from ..client.game_client import GameClient
from loguru import logger




class ImperialPatronage(GameClient):
    
    
    
    async def open_imperial_patronage(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("gdti", {})
            if sync:
                response = await self.wait_for_response("gdti")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
      
      
      
      
      
      
        
        
    async def give_imperial_patronage(
        self,
        devise_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "ddi",
                {
                    "DIV": [{
                        "DII": devise_id, "DIA": amount
                    }]
                }
            )
            if sync:
                response = await self.wait_for_response("ddi")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger





class Attack(GameClient):
    
    
    
    
    async def send_attack(
        self,
        kingdom: int,
        sx: int,
        sy: int,
        tx: int,
        ty: int,
        army: list,
        lord_id: int = 0,
        horses_type: int = -1,
        feathers: int = 0,
        slowdown: int = 0,
        boosters: list = [],
        support_tools: list = [],
        final_wave: list = [],
        sync: bool = True
    ) -> dict | bool:
        
        try:
            
            await self.send_json_message(
                "cra",
                {
                    "SX": sx,
                    "SY": sy,
                    "TX": tx,
                    "TY": ty,
                    "KID": kingdom,
                    "LID": lord_id,
                    "WT": 0,
                    "HBW": horses_type,
                    "BPC": 0,
                    "ATT": 0,
                    "AV": 0,
                    "LP": 0,
                    "FC": 0,
                    "PTT": feathers,
                    "SD": slowdown,
                    "ICA": 0,
                    "CD": 99,
                    "A": army,
                    "BKS": boosters,
                    "AST": support_tools,
                    "RW": final_wave,
                    "ASCT": 0, 
                }
            )
            if sync:
                response = await self.wait_for_response("cra")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger
import random
import asyncio






class Misc(GameClient):
    


    
    async def cooldown(self, min_time: int, max_time: int) -> None:
        
        try:
            
            random_time = random.randint(min_time, max_time)
            await asyncio.sleep(random_time)
            
        except Exception as e:
            logger.error(e)   
        
    


from .core.interpreter import Interpreter
from .castle.castle import Castle
from .castle.defense import Defense
from .map.bookmarks import Bookmarks
from .map.map import Map
from .map.movements import Movements
from .map.attack import Attack
from .map.support import Support
from .map.spy import Spy
from .misc.misc import Misc
from .army.lords import Lords
from .building.buildings import Buildings
from .building.repair import Repair
from .building.buildings_inventory import BuildingsInventory
from .building.wall import Wall
from .building.extension import Extension
from .building.build_items import BuildItems
from .equip.equip import Equip
from .production.soldiers import Soldiers
from .production.tools import Tools
from .production.hospital import Hospital
from .production.refinery import Refinery
from .alliance.alliance import Alliance
from .social.social import Social
from .shops.shop import Shop
from .shops.kings_market import KingsMarket
from .shops.bestseller import BestSeller
from .events.events import Events
from .events.beyond_the_horizon import BeyondTheHorizon
from .events.outer_realms import OuterRealms
from .events.imperial_patronage import ImperialPatronage






 

class GGXS(
    Interpreter,
    Castle,
    Bookmarks,
    Map,
    Attack,
    Movements,
    Misc,
    Lords,
    Defense,
    Soldiers,
    Tools,
    Buildings,
    Repair,
    BuildingsInventory,
    Wall,
    Extension,
    Alliance,
    Equip,
    Social,
    Support,
    Shop,
    KingsMarket,
    BestSeller,
    BuildItems,
    Events,
    BeyondTheHorizon,
    OuterRealms,
    ImperialPatronage,
    Spy,
    Hospital,
    Refinery
):
    pass


from ..client.game_client import GameClient
from loguru import logger





class Refinery(GameClient):
    
    
    
    async def refinery_get_queue(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("crin", {})
            if sync:
                response = await self.wait_for_response("crin")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    async def produce_materials(
        self,
        kingdom: int,
        castle_id: int,
        building_id: int,
        item_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "crst",
                {
                    "KID": kingdom,
                    "AID": castle_id,
                    "OID": building_id,
                    "PWR": 0,
                    "CRID": item_id
                    
                }
            )
            if sync:
                response = await self.wait_for_response("crst")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
            
            
            
    async def cancel_materials_production(
        self,
        kingdom: int,
        castle_id: int,
        building_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "crca",
                {
                    "KID": kingdom,
                    "AID": castle_id,
                    "OID": building_id,
                    "S": 0,
                    "ST": "queue"
                    
                }
            )
            if sync:
                response = await self.wait_for_response("crca")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger






class BestSeller(GameClient):
    
    
    
    async def buy_from_bestseller(
        self,
        bestseller_id: int,
        package_type: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "bso",
                {
                    "OID": package_type,
                    "AMT": amount,
                    "POID": bestseller_id
                }
            )
            if sync:
                response = await self.wait_for_response("bso")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger




class Shop(GameClient):
    
    
    
    async def buy_package_generic(
        self,
        kingdom: int,
        shop_type: int,
        shop_id: int,
        package_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "sbp",
                {
                    "PID": package_id,
                    "BT": shop_type,
                    "TID": shop_id,
                    "AMT": amount,
                    "KID": kingdom,
                    "AID": -1,
                    "PC2": -1,
                    "BA": 0,
                    "PWR": 0,
                    "_PO": -1
                }
            )
            if sync:
                response = await self.wait_for_response("sbp")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    async def buy_from_master_blacksmith(
        self,
        kingdom: int,
        package_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        return await self.buy_package_generic(kingdom, 0, 116, package_id, amount, sync)
    


    async def buy_from_armorer(
        self,
        kingdom: int,
        package_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        return await self.buy_package_generic(kingdom, 0, 27, package_id, amount, sync)
    
    
    async def buy_from_nomad_shop(
        self,
        kingdom: int,
        package_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        return await self.buy_package_generic(kingdom, 0, 94, package_id, amount, sync)
    
    
    
    async def buy_from_blade_coast(
        self,
        kingdom: int,
        package_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        return await self.buy_package_generic(kingdom, 0, 4, package_id, amount, sync)
    
    
    
    async def set_buying_castle(
        self,
        castle_id: int,
        kingdom: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "gbc",
                {
                    "CID": castle_id,
                    "KID": kingdom
                }
            )
            if sync:
                response = await self.wait_for_response("gbc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger





class Extension(GameClient):
    
    
    
    async def buy_extension(
        self, x: int, y: int, rotated: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("ebe", {"X": x, "Y": y, "R": rotated, "CT": 1})
            if sync:
                response = await self.wait_for_response("ebe")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
    async def collect_extension_gift(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("etc", {"OID": building_id})
            if sync:
                response = await self.wait_for_response("etc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger
from typing import Callable, Any
import asyncio




class Social(GameClient):
    
    
    
    
    async def send_player_sms(
        self,
        player_name: str,
        sms_title: str,
        sms_text: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "sms",
                {
                    "RN": player_name,
                    "MH": sms_title,
                    "TXT": sms_text
                }
            )
            if sync:
                response = await self.wait_for_response("sms")
                return response
            return True

        except Exception as e:
            logger.error(e)
            
            
    async def delete_message(self, message_id, sync = True):
        
        try:
            
            await self.send_json_message(
                "dms", {
                    "MID": message_id
                }
            )
            
            if sync:
                response = self.pending_get_status("dms")
                return response
            logger.info(f'Message {message_id} removed!')
            return True
            
        except Exception as e:
            logger.error(e) 
            return False
        
    
    
            
    async def inbox_watcher(
        self,
        handler,
    ) -> asyncio.Task:
        
        watcher_task = asyncio.create_task(
            self.process_data("sne", handler, interval= 0.5)
        )
        return watcher_task
        


from ..client.game_client import GameClient
from loguru import logger





class KingsMarket(GameClient):
    
    
    
    async def start_protection(
        self,
        duration: int,      ### 0:7days, 1: 14 days, 2: 21days, 3: 60 days
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "mps",
                {
                    "CD": duration
                }
            )
            if sync:
                response = await self.wait_for_response("mps")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    async def buy_production_slot(
        self,
        queue_type: int, ## 0 for baracks, 1 for workshop
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "ups",
                {
                    "LID": queue_type
                }
            )
            if sync:
                response = await self.wait_for_response("ups")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False



       
    async def buy_open_gates(
        self,
        kingdom: int,
        castle_id: int,
        duration: int,   ## 0 for 6h, 1 for 12h
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "mos",
                {
                    "CID": castle_id,
                    "KID": kingdom,
                    "CD": duration
                }
            ) 
            if sync:
                response = await self.wait_for_response("mos")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def buy_feast(
        self,
        kingdom: int,
        castle_id: int,
        feast_type: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "bfs",
                {
                    "CID": castle_id,
                    "KID": kingdom,
                    "T": feast_type,
                    "PO": -1,
                    "PWR": 0
                }
            )
            if sync:
                response = await self.wait_for_response("bfs")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


import websockets
import asyncio
import json
import re
from loguru import logger
from typing import Any, Callable, Dict, List, Union, Awaitable
import uuid
from .memcache import GGSCache
import inspect
import aiohttp

HandlerType = Callable[[Any], Union[Any, Awaitable[Any]]]
version_url = "https://empire-html5.goodgamestudios.com/default/items/ItemsVersion.properties"




class GameClient:
    
    def __init__(self, url: str, server_header: str, username: str, password: str):
        
        self.url = url
        self.server_header = server_header
        self.username = username
        self.password = password
        self._pending = GGSCache()
        self.connected = asyncio.Event()
        
        
    
    
    
    
        
    async def connect(self) -> None:
        async with websockets.connect(self.url) as ws:
            self.ws = ws
            self.connected.set()
            listener_task = asyncio.create_task(self.listen_messages())
            keep_alive_task = asyncio.create_task(self.keep_alive())
            connected = await self.init()
            if not connected:
                await self.disconnect()
            
            
            runner_task = asyncio.create_task(self.run_jobs())
            await asyncio.gather(listener_task, runner_task, keep_alive_task)




    async def init(self) -> bool:
        await self.init_socket()
        return await self.login(self.username, self.password)



    async def run_jobs(self):
        pass # va fi suprapusa cu o functie custom 
    
    
    
    
    async def send(self, message):
        
        try:
            await self.ws.send(message)
        except Exception as e:
            logger.error(f"Error occured: {e}")    
       
           
            
    async def send_message(self, parts: List[str]) -> None:
        await self.send("%".join(["", *parts, ""]))



    async def send_raw_message(self, command: str, data: List[Any]) -> None:
        json_parts = [json.dumps(item) if isinstance(item, (dict, list)) else item for item in data]
        await self.send_message(["xt", self.server_header, command, "1", *json_parts])



    async def send_json_message(self, command: str, data: Dict[str, Any]) -> None:
        await self.send_message(["xt", self.server_header, command, "1", json.dumps(data)])



    async def send_xml_message(self, t: str, action: str, r: str, data: str) -> None:
        await self.send(f"<msg t='{t}'><body action='{action}' r='{r}'>{data}</body></msg>") 





    async def listen_messages(self):
        try:
            async for message in self.ws:
                await self.handle_message(message)
        except websockets.ConnectionClosed:
            logger.error("Connection closed. Reconnecting...")
            await asyncio.sleep(5)
            await self.connect()




    async def handle_message(self, message):
        if isinstance(message, bytes):
            message = message.decode('utf-8')
            self._message_parser(message)
            
            




    def _message_parser(self, message: str, caching: bool = True) -> Dict[str, Any]:
        
        if message.startswith("<"):
            
            xml_pattern = r"<msg t='(.*?)'><body action='(.*?)' r='(.*?)'>(.*?)</body></msg>"
            m = re.search(xml_pattern, message)
            t_value, action_value, r_value, data_value = m.groups()
            
            
            msg = {
                "type": "xml",
                "payload": {
                    "t": t_value,
                    "action": action_value,
                    "r": int(r_value),
                    "data": data_value
                }     
            }
             
            if caching:
                self._add_pending('xml', msg)
                
            return msg
          
            
            
        else:
            
            message = message.strip("%").split("%")
            message = {
                "type": "json",
                "payload": {
                    "command": message[1],
                    "status": int(message[3]),
                    "data": "%".join(message[4:]) if len(message) > 4 else None
                }
            }
            if message["payload"]["data"] and message["payload"]["data"].startswith("{"):
                message["payload"]["data"] = json.loads(message["payload"]["data"])
            
            if caching:
                self._add_pending('json', message)
             
            return message
                

    def _add_pending(self, type: str, parsed_message: Dict[str, Any]) -> None:
        
        if type == "json":
            
            try:

                key_uuid = f'{parsed_message["payload"]["command"]}_{str(uuid.uuid4())}'
                self._pending.set_with_ttl(key_uuid, parsed_message, 5.0)
                
            
            except Exception as e:
                logger.error(e)
        
        
        else:
            
            try:
                key_uuid = f'{parsed_message["payload"]["action"]}_{str(uuid.uuid4())}'
                self._pending.set_with_ttl(key_uuid, parsed_message, 5.0)
                

            except Exception as e:
                logger.error(e)



    async def disconnect(self) -> None:
        
        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
        
        for task in tasks:
            task.cancel()
            
        await asyncio.gather(*tasks, return_exceptions=True)
        
        try:
            await self.ws.close()
            logger.info("Connection closed succesfully!")
        
        except Exception as e:
            logger.error(f"Error occured: {e}")        

    
    async def pending_get_status(self,
                                 prefix: str,
                                 status: int = 0,
                                 timeout: float = 5.0,
                                 interval: float = 0.5) -> bool:
        
        start = asyncio.get_event_loop().time()
        while (asyncio.get_event_loop().time() - start) < timeout:
            result_status = self._pending.get_key_status(prefix)
            if result_status is not None:
                return result_status == status
            await asyncio.sleep(interval)
        
        return False
                
    async def pending_match_xml(self,
                                prefix: str,
                                expected: dict,
                                timeout: float = 5.0,
                                interval: float = 0.1) -> bool:
        
        start = asyncio.get_event_loop().time()
        while (asyncio.get_event_loop().time() - start) < timeout:
            pending_entries = self._pending.find_by_prefix(prefix)
            for msg in pending_entries.values():
                if isinstance(msg, dict):
                    payload = msg.get("payload", {})
                    if payload == expected:
                        return True
            await asyncio.sleep(interval)
            
        return False                       

     
        
    async def init_socket(self):
        
        try:
            await self.send_xml_message("sys", "verChk", "0", "<ver v='166' />")
            await self.send_xml_message("sys", "login", "0", 
                                        f"<login z='{self.server_header}'><nick><![CDATA[]]></nick><pword><![CDATA[1123010%fr%0]]></pword></login>")
            await self.send_xml_message("sys", "autoJoin", "-1", "")
            await self.send_xml_message("sys", "roundTrip", "1", "")
            
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    
    
    
    
    
    
    
    async def ping(self):
        
        try:
            await self.send_raw_message("pin", ["<RoundHouseKick>"])
            return True
        
        except Exception as e:
            logger.error(e)
            return False


    async def keep_alive(self, interval: int = 60):
        
        await self.connected.wait()
        while self.connected.is_set():
            try:
                await asyncio.sleep(interval)
                await self.ping()
            except asyncio.CancelledError:
                logger.info("keep_alive task has been canceled!.")
        
            except Exception as e:
                logger.error(f"keep_alive error: {e}")
                raise
                
          



    async def process_data(
        self,
        prefix: str,
        handler: HandlerType,
        interval: float = 0.5
    ) -> None:
        
        while self.connected.is_set():
            key_map = self._pending.get_keys_data(prefix)
            for key, data in list(key_map.items()):
                try:
                    result = handler(data)
                    if inspect.isawaitable(result):
                        await result
        
                except Exception as e:
                    logger.error(e)

                finally:
                    self._pending.pop(key, None)
            await asyncio.sleep(interval)




    async def wait_for_response(self,
                                prefix: str,
                                interval: float = 0.5,
                                timeout: float = 5.0):
        
        start = asyncio.get_event_loop().time()
        while (asyncio.get_event_loop().time() - start) < timeout:
            result_data = self._pending.get_data_by_status(prefix, 0)
            if result_data is not None:
                return result_data
            await asyncio.sleep(interval)
        
        return None    
        
    
    
    async def process_response_status(
        self,
        prefix: str,
        expected_status: int,
        interval: float = 0.5,
        timeout: float = 5.0
    ):
        
        start = asyncio.get_event_loop().time()
        while (asyncio.get_event_loop().time() - start) < timeout:
            result_data = self._pending.process_by_status(prefix, expected_status)
            if result_data is not None:
                return result_data
            await asyncio.sleep(interval)
        
        return None   
            
    
    
    
    
    
    
    
    
    async def login(
        self,
        username: str,
        password: str
    ) -> bool:
        
        if not self.connected.is_set():
            logger.error("Not connected yet!")
            return False
        self._pending.clear()   
            
        try:
            await self.send_json_message(
                "lli",
                {
                    "CONM": 175,
                    "RTM": 24,
                    "ID": 0,
                    "PL": 1,
                    "NOM": username,
                    "PW": password,
                    "LT": None,
                    "LANG": "fr",
                    "DID": "0",
                    "AID": "1674256959939529708",
                    "KID": "",
                    "REF": "https://empire.goodgamestudios.com",
                    "GCI": "",
                    "SID": 9,
                    "PLFID": 1
                },
            )
            response_status = await self.process_response_status("lli", 0, 0.3, 5.0)
            if isinstance(response_status, bool):
                if response_status:
                    return response_status
                else:
                    logger.error("Login failed, check your details!")
                    return False
                
            if isinstance(response_status, dict):
                cd_value = response_status["CD"]
                logger.debug(f'Connection locked by the server! Reconnect in {cd_value} sec!')
                await asyncio.sleep(cd_value)
                return await self.login(username, password)
                    
        except Exception as e:
            logger.error(e) 
            return False
        
        logger.error("Unexpected response in login status!")
        return False
    
    
    async def fetch_game_db(self) -> dict:
        
        async with aiohttp.ClientSession() as session:
            async with session.get(version_url) as resp:
                resp.raise_for_status()
                text = await resp.text()
                _, version = text.strip().split("=", 1)
                version = version.strip()
            
            db_url = f"https://empire-html5.goodgamestudios.com/default/items/items_v{version}.json"
            async with session.get(db_url) as db_resp:
                db_resp.raise_for_status()
                data = await db_resp.json()
                return data
        


from ..client.game_client import GameClient
from loguru import logger






class Soldiers(GameClient):
    
    
    
    async def get_recruitment_queue(self, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("spl", {"LID": 0})
            if sync:
                response = await self.wait_for_response("spl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
     
        
    async def recruit_soldiers(
        self,
        castle_id: int,
        wod_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            
            await self.send_json_message(
                "bup",
                {
                    "LID": 0,
                    "WID": wod_id,
                    "AMT": amount,
                    "PO": -1,
                    "PWR": 0,
                    "SK": 73,
                    "SID": 0,
                    "AID": castle_id
                }
            )
            if sync:
                response = await self.wait_for_response("bup")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False
        
        
        
        
    async def cancel_recruitment(self, slot_type: str, slot: int, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("mcu", {"LID": 0, "S": slot, "ST": slot_type})
            if sync:
                response = await self.wait_for_response("mcu")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def recruitment_alliance_help(self, sync: bool = True) -> dict | bool:
        
        try:
            await self.send_json_message("ahr", {"ID": 0, "T": 6})
            if sync:
                response = await self.wait_for_response("ahr")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
        
        
    async def get_units_inventory(
        self,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("gui", {})
            if sync:
                response = await self.wait_for_response("gui")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False        
        
        
    
    async def delete_units(
        self,
        wod_id: int,
        amount: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "dup",
                {
                    "WID": wod_id,
                    "A": amount,
                    "S": 0
                }
            )
            if sync:
                response = await self.wait_for_response("dup")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False   


from loguru import logger






class Utils:
    
    def __init__(self):
        pass
    
    
    
    def skip_calculator(
        self, time: int, skip_type: list[str] = None
        ) -> list[str]:
        
        
        all_skip_values = [(86400, "MS7"), (18000, "MS6"), (3600, "MS5"), (1800, "MS4"), (600, "MS3"), (300, "MS2"), (60, "MS1")]
        
        if skip_type:
            skip_values = [(sec, lbl) for sec, lbl in all_skip_values if lbl in skip_type]
            if not skip_values:
                logger.error("Your skips are not allowed!")
                
        else:
            skip_values = all_skip_values
            
        minutes = time // 60
        skip_minutes = [(sec // 60, label) for sec, label in all_skip_values]
        INF = float('inf')
        dp = [INF] * (minutes + 1)
        prev = [None] * (minutes + 1)
        dp[0] = 0
        
        for i in range(1, minutes + 1):
            for m, label in skip_minutes:
                if i >= m and dp[i - m] + 1 < dp[i]:
                    dp[i] = dp[i - m] + 1
                    prev[i] = (i - m, label)
        skips = []
        cur = minutes
        while cur > 0 and prev[cur]:
            j, label = prev[cur]
            skips.append(label)
            cur = j
        used = sum(sec * skips.count(lbl) for sec, lbl in all_skip_values)
        rem = time - used
        if rem > 0:
            for sec, label in sorted(all_skip_values, key=lambda x: x[0]):
                if sec >= rem:
                    skips.append(label)
                    break
        
        return skips
    
    
    
    


from ..client.game_client import GameClient
from loguru import logger


class Bookmarks(GameClient):
    
    
    async def get_bookmarks(self, sync: bool = True) -> dict:
        
        try:
            
            await self.send_json_message("gbl", {})
            if sync:
                response = await self.wait_for_response("gbl")
                return response
            return True
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger
import asyncio
from ..utils.utils import Utils




class Buildings(GameClient):
    
    
    
    async def build(
        self,
        wod_id: int,
        x: int,
        y: int,
        rotated: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "ebu",
                {
                    "WID": wod_id,
                    "X": x,
                    "Y": y,
                    "R": rotated,
                    "PWR": 0,
                    "PO": -1,
                    "DOID": -1,
                }
            )
            if sync:
                response = await self.wait_for_response("ebu")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
            
            
    async def upgrade_building(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("eup", {"OID": building_id, "PWR": 0, "PO": -1})
            if sync:
                response = await self.wait_for_response("eup")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
     
     
    async def move_building(
        self,
        building_id: int,
        x: int,
        y: int,
        rotated: int = 0,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "emo", {"OID": building_id, "X": x, "Y": y, "R": rotated}
            )
            if sync:
                response = await self.wait_for_response("emo") 
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            
            
    async def sell_building(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("sbd", {"OID": building_id})
            if sync:
                response = await self.wait_for_response("sbd") 
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    async def destroy_building(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("edo", {"OID": building_id})
            if sync:
                response = await self.wait_for_response("edo") 
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
            

    
    
    
    async def skip_construction_free(
        self,
        building_id: int,
        free_skip: int = 1,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("fco", {"OID": building_id, "FS": free_skip})
            if sync:
                response = await self.wait_for_response("fco") 
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
            
            
    async def time_skip_construction(
        self,
        building_id: int,
        time_skip: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("msb", {"OID": building_id, "MST": time_skip})
            if sync:
                response = await self.wait_for_response("msb") 
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    async def wait_finish_construction(
        self, building_id: int, timeout: int
    ) -> dict | bool:
        
        try:
            response = await self.wait_for_response(
                "fbe", {"OID": building_id}, timeout=timeout
            )
            return response
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    
            
    async def instant_build(
        self,
        building_id: int,
        wod_id: int,
        x: int,
        y: int,
        rotated: int = 0,
        time_skips: list[str] = [],
        cooldown: int = 0,
        free_skip: int = 1,
        sync: bool = True
    ) -> None:
        
        try:
            await self.build(wod_id, x, y, rotated, sync=sync)
            for skip in time_skips:
                await self.time_skip_construction(building_id, skip, sync=sync)   
            asyncio.sleep(cooldown)
            await self.skip_construction_free(building_id, free_skip, sync=sync)
            if sync:
                await self.wait_finish_construction(building_id)
        
        except Exception as e:
            logger.error(e)
        
       
    async def instant_upgrade(
        self,
        building_id: int,
        time_skips: list[str] = [],
        cooldown: int = 0,
        free_skip: int = 1,
        sync: bool = True
    ) -> None:
        
        try:
            await self.upgrade_building(building_id, sync=sync) 
            for skip in time_skips or []:
                await self.time_skip_construction(building_id, skip, sync=sync)
            asyncio.sleep(cooldown)
            await self.skip_construction_free(building_id, free_skip, sync=sync)
            if sync:
                await self.wait_finish_construction(building_id)
        
        except Exception as e:
            logger.error(e)
            
    
    async def instant_destroy(
        self,
        building_id: int,
        time_skips: list[str] = [],
        cooldown: int = 0,
        free_skip: int = 1,
        sync: bool = True
    ) -> None:
        
        try:
            await self.destroy_building(building_id, sync=sync) 
            for skip in time_skips or []:
                await self.time_skip_construction(building_id, skip, sync=sync)
            asyncio.sleep(cooldown)
            await self.skip_construction_free(building_id, free_skip, sync=sync)
            
        
        except Exception as e:
            logger.error(e)
            
            
   



from ..client.game_client import GameClient
from loguru import logger





class Support(GameClient):
    
    
    
    
    
    
    async def get_support_info(
        self,
        sx: int,
        sy: int,
        tx: int,
        ty: int,
        sync: bool = True 
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "sdi",
                {
                    "TX": tx,
                    "TY": ty,
                    "SX": sx,
                    "SY": sy
                }
            )
            if sync:
                response = await self.wait_for_response("sdi")
                return response
            return True

        except Exception as e:
            logger.error(e)
            return False
        
                
    
    
    
    
    
    async def send_support(
        self,
        units: list,
        sender_id: int,
        tx: int,
        ty: int,
        lord_id: int,
        camp_time: int = 12,
        horses_type: int = -1,
        feathers: int = 0,
        slowdown: int = 0,
        sync: bool = True 
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "cds",
                {
                    "SID": sender_id,
                    "TX": tx,
                    "TY": ty,
                    "LID": lord_id,
                    "WT": camp_time,
                    "HBW": horses_type,
                    "BPC": 0,
                    "PTT": feathers,
                    "SD": slowdown,
                    "A": units
                }
            )
            if sync:
                response = await self.wait_for_response("cds")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
from loguru import logger
import asyncio



class Movements(GameClient):
    
    
    async def get_movements(self, sync = True) -> dict | bool:
        
        try:
            
            await self.send_json_message("gam", {})
            
            if sync:
                response = await self.wait_for_response("gam")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


    async def tower_watch(
        self,
        handler,
    ) -> asyncio.Task:
        
        tower_task = asyncio.create_task(
            self.process_data("gam", handler, interval= 0.5)
        )
        return tower_task
        
        
    
    


from ..client.game_client import GameClient
from loguru import logger







class Equip(GameClient):
    
    
    
    
    async def get_equip_inventory(
        self, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("gei", {})
            if sync:
                response = await self.wait_for_response("gei")
                return response
            return True
        
        except Exception as e:
            logger.error(e) 
            return False
        
    
    async def remove_equip(
        self, equip_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "seq",
                {
                    "EID": equip_id,
                    "LID":-1,
                    "EX":0,
                    "LFID":-1
                }
            )
            if sync:
                response = await self.wait_for_response("nrf")
                return response
            return True
        
        except Exception as e:
            logger.error(e) 
            return False
        
    
    async def get_gems_inventory(
        self, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("ggm", {})
            if sync:
                response = await self.wait_for_response("ggm")
                return response
            return True

        except Exception as e:
            logger.error(e) 
            return False
    
    
    
        
    async def remove_gem(self, gem_id: int, sync: bool = True) -> dict | bool:
        
        try:
            
            await self.send_json_message(
                "sge",
                {
                    "GID": gem_id,
                    "RGEM": 0,
                    "LFID": -1
                }
            )
            if sync:
                response = await self.pending_get_status("sge")
                return response
            return True  
            
        except Exception as e:
            logger.error(e) 
            return False
        
        
        
    async def gem_remover(
        self, sync: bool = True
    ) -> None:
        
        gems_inventory = await self.get_gems_inventory()
        gem_data = gems_inventory["GEM"]
        for gem in gem_data:
            gem_id = gem[0]
            gem_qty = gem[1]
            for _ in range(gem_qty):
                
                try:
                    await self.remove_gem(gem_id, sync=sync)
                
                except Exception as e:
                    logger.error(e)
            
        logger.info("All useless gems has been removed")
        
        
        
        
    async def old_equip_remover(
        self, sync: bool = True
    ) -> None:
        
        useless = [1, 2, 3, 11, 12, 13]
        equip_data = await self.get_equip_inventory()
        equip_obj = equip_data["I"]
        for equip in equip_obj:
            if equip[3] in useless and equip[4] > 0:
                
                try:
                    await self.remove_equip(equip_id=equip[0], sync=sync)
                
                except Exception as e:
                    logger.error(e)
                    return None
        
        logger.info("All useless equipements has been removed")


from ..client.game_client import GameClient
from loguru import logger





class Wall(GameClient):
    
    
    
    async def upgrade_wall(
        self, building_id: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("eud", {"OID": building_id, "PWR": 0, "PO": -1})
            if sync:
                response = await self.wait_for_response("eud")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False


from ..client.game_client import GameClient
import asyncio
from typing import Callable, Dict







class Interpreter(GameClient):
    
    
    async def _interpreter(self, action: Dict[str, Callable]) -> None:
        
        
        auto_tasks = []
        for prefix, handler in action.items():
            task = asyncio.create_task(self.process_data(prefix, handler, 0.5))
            auto_tasks.append(task)
            
        await asyncio.gather(*auto_tasks)  
            
            
    



from collections.abc import MutableMapping
from typing import Any, Callable, Optional, Dict, Union
import asyncio




class GGSCache(MutableMapping):
    
    def __init__(
        self,
        default_ttl: float = 5.0,
        on_expire: Optional[Callable[[Any, Any], None]] = None
    ):
        self._default_ttl = default_ttl
        self._on_expire = on_expire
        
        self._data: Dict[Any, Any] = {}
        self._tasks: Dict[Any, asyncio.Task] = {}
        
    
    def __getitem__(self, key: Any):
        return self._data[key]
    
    
    def __setitem__(self, key: Any, value: Any) -> None:
        
        self.set_with_ttl(key, value, self._default_ttl)
    
    
    def __len__(self) -> int:
        return len(self._data)
    
    def __iter__(self):
        return iter(self._data)
    
    
        
    def set_with_ttl(self, key: Any, value: Any, ttl: float) -> None:
        
        if key in self._tasks:
            self._tasks[key].cancel()
        self._data[key] = value
        self._tasks[key] = asyncio.create_task(self._expire_later(key, ttl))
 
    
    def __delitem__(self, key):
        
        if key in self._tasks:
            self._tasks[key].cancel()
            del self._tasks[key]
 
        del self._data[key]
 
    async def _expire_later(self, key: Any, ttl: float) -> None:
        
        try:
            await asyncio.sleep(ttl)
            val = self._data.pop(key, None)
            self._tasks.pop(key, None)
            if self._on_expire and val is not None:
                self._on_expire(key, val)
 
        except asyncio.CancelledError:
            return
 
 
    def find_by_prefix(self, prefix: str) -> Dict[Any, Any]:
        
        result: Dict[Any, Any] = {}
        for k, v in self._data.items():
            if str(k).startswith(prefix):
                result[k] = v
        
        return result

        
    def find_by_value(self, target: Any) -> list[Any]:
        return [k for k, v in self._data.items() if v == target]
    
    
    def clear(self) -> None:
        for t in self._tasks.values():
            t.cancel()
        
        self._data.clear()
        self._tasks.clear()
        
        
    def get_keys_data(self, prefix: str) -> Dict[Any, Any]:
        
        result: Dict[Any, Any] = {}
        for key, val in self._data.items():
            if str(key).startswith(prefix):
                if not isinstance(val, dict):
                    result[key] = None
                    continue
            
                payload = val.get("payload")
                if not isinstance(payload, dict):
                    result[key] = None
                    continue
            
                result[key] = payload.get("data")
        
        return result
        
        
    def get_key_status(self, prefix: str) -> Optional[int]:
        
        for key, val in self._data.items():
            if str(key).startswith(prefix):
                if not isinstance(val, dict):
                    return None
                payload = val.get("payload")
                if not isinstance(payload, dict):
                    return None
                status = payload.get("status")
                return status if isinstance(status, int) else None
        
        return None
    
    def get_data_by_status(self, prefix: str, expected_status: int = 0) -> Dict[Any, Any]:
        
        for key, val in self._data.items():
            if str(key).startswith(prefix):
                if not isinstance(val, dict):
                    return None
                payload = val.get("payload")
                if not isinstance(payload, dict):
                    return None
                status = payload.get("status")
                data = payload.get("data")
                return data if status == expected_status else None
        
        return None
    
    
    ## special function for game
    
    def process_by_status(
        self,
        prefix: str, 
        expected_status: int = 0               
        ) -> Union[Dict[Any, Any], bool]:
        
        for key, val in self._data.items():
            if str(key).startswith(prefix):
                if not isinstance(val, dict):
                    return None
                payload = val.get("payload")
                if not isinstance(payload, dict):
                    return None
                status = payload.get("status")
                data = payload.get("data")
                
                if status == expected_status:
                    return True
                
                if status == 20:
                    return False
                
                if status == 453:
                    return data
        
        return None
        
        
        


from ..client.game_client import GameClient
from loguru import logger







class BuildItems(GameClient):
    
    
    
    
    
    async def equip_build_item(
        self,
        kingdom_id: int,
        castle_id: int,
        building_id: int,
        slot_id: int,
        item_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        
        try:
            await self.send_json_message(
                "rpc",
                {
                    "OID": building_id,
                    "CID": item_id,
                    "SID": slot_id,
                    "M": 0,
                    "KID": kingdom_id,
                    "AID": castle_id
                }
            )
            if sync:
                response = await self.wait_for_response("rpc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    
    
    
    
    
    
    


from ..client.game_client import GameClient
from loguru import logger






class Alliance(GameClient):
    
    
    
    async def get_chat(
        self, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("acl", {})
            if sync:
                response = await self.wait_for_response("acl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    async def write_on_chat(
        self,
        message: str,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "acm",
                {
                    "M": message
                }
            )
            if sync:
                response = await self.wait_for_response("acm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
    
    
    
       
        
    async def help_alliance_member(
        self, kingdom: int, help_id: int, sync: bool = True
    ) -> dict | bool:

        try:
            await self.send_json_message("ahc", {"LID": help_id, "KID": kingdom})
            if sync:
                response = await self.wait_for_response("ahc")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False
        
    
    
    
       
    async def help_alliance_all(
        self, kingdom: int, sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message("aha", {"KID": kingdom})
            if sync:
                response = await self.wait_for_response("aha")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False  
        
        
    async def invite_player(
        self,
        user_id: int,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "aip",
                {
                    "SV": user_id
                }
            )
            if sync:
                response = await self.wait_for_response("aip")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
        
        
    async def rank_player(
        self,
        account_id: int,
        rank: int,   ## 0-8 leader to member
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "arm",
                {
                    "PID": account_id,
                    "R": rank
                }  
            )
            if sync:
                response = await self.wait_for_response("arm")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 
        
      
      
        
        
    async def mass_message(
        self,
        text: str,
        title: str = None,
        sync: bool = True
    ) -> dict | bool:
        
        try:
            await self.send_json_message(
                "anl",
                {
                    "SJ": title,
                    "TXT": text
                }
            )
            if sync:
                response = await self.wait_for_response("anl")
                return response
            return True
        
        except Exception as e:
            logger.error(e)
            return False 